/*
*	buffer
*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "buffer.h"

/*
*	create a new buffer
*/
buf_t	*create_buf()
{
	buf_t	*buf;

	buf = malloc(sizeof(buf_t));

	buf->size  = 1024;
	buf->len   = 0;
	buf->ptr   = malloc(buf->size);
	
	buf->pos   = 0;

	memset(buf->ptr, 0, buf->size);
	return buf;
}

/*
*	destroy a buffer
*/
void	destroy_buf(buf_t *buf)
{
	if	( buf )	{
	 	free(buf->ptr);
		free(buf);
		}
}

/*
*	resize buffer
*/
void	resize_buf(buf_t *buf, int size)
{
	buf->ptr = realloc(buf->ptr, size);
	memset(buf->ptr+buf->size, 0, size-buf->size);
	buf->size = size;
}

/*
*	check and resize
*/
void	check_for_resize(buf_t *buf, int rqsize)
{
	if	( rqsize >= buf->size ) 
		resize_buf(buf, buf->size+1024);
}

/*
*	insert a character at the current position
*/
void	insert_char(buf_t *buf, int ch)
{
	check_for_resize(buf, buf->len+1);

	if	( buf->pos >= buf->len )	{
		/* add to the end */
		buf->ptr[buf->pos] = ch;
		}
	else	{
		/* insert char */
		char	*r;

		r = strdup(buf->ptr+buf->pos);
		buf->ptr[buf->pos] = ch;
		strcpy(buf->ptr+buf->pos+1, r);
		free(r);
		}

	buf->len ++;
	buf->pos ++;
}

/*
*	overwrite
*/
void	store_char(buf_t *buf, int ch)
{
	check_for_resize(buf, buf->len+1);

	buf->ptr[buf->pos] = ch;
	buf->pos ++;
}

/*
*	returns the current line
*/
int		cur_line(buf_t *buf)
{
	int		ln = 0;
	char	*p = buf->ptr;

	while ( p != buf->ptr+buf->pos )	{
		if	( *p == '\n' )
			ln ++;
		p ++;
		}
	return ln;
}

/*
*	returns the current column
*/
int		cur_col(buf_t *buf)
{
	int		i, c = 0;

	for ( i = buf->pos; i >= 0; i --, c ++ )	{
		if	( buf->ptr[i] == '\n' )	{
			break;
			}
		}
	return c-1;
}

/*
*	delete one char back (backspace)
*/
void	del_one_char_back(buf_t *buf)
{
	if	( buf->pos )	{
		if	( buf->ptr[buf->pos] )	
			memcpy(buf->ptr+(buf->pos-1), buf->ptr+buf->pos, buf->len-buf->pos);

		buf->pos --;
		buf->len --;
		buf->ptr[buf->len] = '\0';
		}
}

/*
*	delete one char (delete)
*/
void	del_one_char(buf_t *buf)
{
	if	( buf->pos < buf->len )	{
		char	*r;

		r = strdup(buf->ptr+buf->pos+1);
		strcpy(buf->ptr+buf->pos, r);
		free(r);
		}
}

/*
*	set display window
*/
void	set_bufwin(buf_t *buf, int x1, int y1, int x2, int y2)
{
	buf->wx1 = x1;	buf->wx2 = x2;
	buf->wy1 = y1;	buf->wy2 = y2;
}

/*
*	display buffer
*/
void	display_buf(buf_t *buf)
{
	printf("\033[2J\033[1;1H");
	printf("%s", buf->ptr);
	fflush(stdout);
}
